package	jp.co.sds_corp.util.db;

import java.sql.Connection;
import java.sql.Driver;
import java.sql.DriverManager;
import java.sql.SQLException;
import java.util.Date;
import java.util.Hashtable;
import java.util.Map;
import java.util.Properties;
import java.util.ResourceBundle;
import java.util.concurrent.ConcurrentHashMap;

import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.NamingException;
import javax.sql.DataSource;

import jp.co.sds_corp.util.common.JarClassLoader;
import jp.co.sds_corp.util.config.IConfiguration;
import jp.co.sds_corp.util.log.CmnLogFactory;
import jp.co.sds_corp.util.log.CmnLogger;

/**
* JDBC接続情報を基に、Connectionを生成して返します。
* このとき生成されるConnectionは、DbConnectionでラップされます。
* @see	DbInfo
* @see	DbConnection
*/
public class DbConnectionFactory
{
	static final CmnLogger logger = CmnLogFactory.getLogger();
	/** SQL文出力フラグ */
	static boolean	sqlOutput = true;
	Context context = null;
	Map<String,DataSource> mapDS = new ConcurrentHashMap<String,DataSource>();

	static DbConnectionFactory base;
	static {
		init();
	}

	public static void init() {
		init(new DbConnectionFactory());
	}
	public static void init(DbConnectionFactory base) {
		DbConnectionFactory.base = base;
	}

	/**
	* SQL文出力フラグを設定します。
	* @param	var		SQL文出力フラグ
	*/
	public static void setSqlOutput( boolean var ) {
		sqlOutput = var;
	}
	/**
	* SQL文出力フラグを取得します。
	* @return	SQL文出力フラグ
	*/
	public static boolean isSqlOutput() {
		return sqlOutput;
	}

	/**
	* 指定されたpropertiesファイルで初期化します。
	* @param	prop	properiesファイル
	*/
	public static DbInfo getDbInfo( ResourceBundle prop ) throws DbException {
		DbInfo inf = new DbInfo( prop );
		return inf;
	}

	/**
	* コンストラクタ
	* 指定されたキーによりＤＢの接続情報を取得する
	* @param	key		このキーでprpertiesファイルを検索する
	*/
	public static DbInfo getDbInfo( String key ) throws DbException {
		DbInfo inf = new DbInfo( key );
		return inf;
	}

	/**
	* 指定されたXMLで初期化します。
	* @param	config	DB接続情報
	*/
	public static DbInfo getDbInfo( IConfiguration config ) throws Exception {
		DbInfo inf = new DbInfo( config );
		return inf;
	}

	/**
	* ＤＢの接続情報により、ＤＢにコネクトする
	*/
	public static Connection getConnection(DbInfo inf) throws DbException {
		return base.getConnectionImpl(inf);
	}

	protected Connection getConnectionImpl( DbInfo inf ) throws DbException {
		Properties prop = inf.getProperties();
		Connection con = null;
		if ( isSqlOutput()) {
			logger.sql( this, inf.toString() );
		}
		Date start = new Date();
		try {
			// JDBCドライバのセット
			if ( inf.getType() == DbInfo.Type.DATASOURCE ) {
				DataSource ds = resolveDataSource( inf );
				if ( inf.getUser() != null && inf.getUser().length() > 0 ) {
					String pwd = inf.getPassword();
					con = ds.getConnection( inf.getUser(), pwd );
				} else {
					con = ds.getConnection();
				}
			} else
			if ( inf.getType() == DbInfo.Type.JDBC ) {
				Driver drv = null;
				if ( inf.getJar() != null && inf.getJar().length() > 0 ) {
					JarClassLoader loader =
						new JarClassLoader( inf.getJar(),
							inf.getClass().getClassLoader() );
					Class<?> cl = loader.loadClass( inf.getDriver() );
					drv = (java.sql.Driver )cl.newInstance();
					return getConnection( drv, inf );
				} else {
					drv = (Driver )Class.forName( inf.getDriver() ).newInstance();
				}
				DriverManager.registerDriver( drv );
				if ( inf.getUser() != null && inf.getUser().length() > 0 ) {
					prop.put( "user", inf.getUser() );
					if ( inf.getPassword() != null && inf.getPassword().length() > 0 ) {
						String pwd = inf.getPassword();
						if ( inf.isEncrypted()) {
							pwd = PasswordEncrypt.decrypt(
									inf.getEncryptKey(), pwd );
						}
						prop.put( "password", pwd );
					}
				}
				con = DriverManager.getConnection( inf.getUrl(), prop );
			}
		} catch ( SQLException ex ) {
			throw new DbException( "connect() error", inf.toString(), ex );
		} catch ( Exception ex ) {
			throw new DbException( "connect() error", inf.getDriver(), ex );
		}

		DbConnection ret = new DbConnection( con );
		ret.setSqlOutput( isSqlOutput() );
		Date now = new Date();
		long time = now.getTime() - start.getTime();
		logger.sql( this, "DB connect elapsed time " + time + " ms.");
		return ret;
	}

	DataSource resolveDataSource( DbInfo inf ) throws NamingException {
		DataSource source = null;
		String name = inf.getDataSource();
		source = mapDS.get( name );
		if ( source != null ) {
			return source;
		}

		if ( inf.getContext() != null && inf.getContext().length() > 0 ) {
			Hashtable<String,String> ht = new Hashtable<String,String>();
			ht.put( Context.INITIAL_CONTEXT_FACTORY, inf.getContext() );
			if ( inf.getJndiUrl() != null ) {
				ht.put( Context.PROVIDER_URL, inf.getJndiUrl() );
			}
			Context ctx = new InitialContext( ht );
			source = (DataSource )ctx.lookup( name );
			return source;
		}
		if ( context == null ) {
			context = new InitialContext();
		}
		source = (DataSource )context.lookup( name );
		mapDS.put( name, source );
		return source;
	}

	/**
	* ドライバを指定してDBサーバと接続します。
	* @param	drv		ドライバ
	*/
	public static Connection getConnection( Driver drv, DbInfo inf ) throws DbException {
		return base.getConnectionImpl(drv, inf);
	}

	protected Connection getConnectionImpl( Driver drv, DbInfo inf ) throws DbException {
		Properties prop = inf.getProperties();
		Connection con;
		if ( isSqlOutput()) {
			logger.sql( this, inf.toString() );
		}
		Date start = new Date();
		try {
			DriverManager.registerDriver( drv );
			if ( inf.getUser() != null && inf.getUser().length() > 0 ) {
				prop.put( "user", inf.getUser() );
				if ( inf.getPassword() != null && inf.getPassword().length() > 0 ) {
					String pwd = inf.getPassword();
					if ( inf.isEncrypted()) {
						pwd = PasswordEncrypt.decrypt(
								inf.getEncryptKey(), pwd );
					}
					prop.put( "password", pwd );
				}
			}
			con = drv.connect( inf.getUrl(), prop );
		} catch ( SQLException ex ) {
			throw new DbException( "connect() error", inf.toString(), ex );
		} catch ( Exception ex ) {
			throw new DbException( "connect() error", inf.getDriver(), ex );
		}

		DbConnection ret = new DbConnection( con );
		ret.setSqlOutput( isSqlOutput() );
		Date now = new Date();
		long time = now.getTime() - start.getTime();
		logger.sql( this, "DB connect elapsed time " + time + " ms.");
		return ret;
	}

}
