package jp.co.sds_corp.util.math;

import java.util.Random;

import jp.co.sds_corp.util.log.CmnLogFactory;
import jp.co.sds_corp.util.log.CmnLogger;

/**
* ゴールシークによる最適解の算出
*/
public class GoalSeeking
{
	static final CmnLogger logger = CmnLogFactory.getLogger();

	/**
	* 関数f(x)
	*/
	public interface Function {
		public double of(double x) throws UtilMathException;
	}

	public static final double EPS          = 1.0e-4;
	public static final int PLUS_DIRECTION  = 0;
	public static final int MINUS_DIRECTION = 1;
	public static Random rand;

	/**
	* ゴールシークの実行
	* @param	f		計算処理関数
	* @param	s		初期値
	* @param	t		目標値
	* @param	d		方向
	* @param	loop	反復回数
	* @return	結果値
	*/
	public static double solve(Function f, double s, double t, int d, int loop)
		throws UtilMathException
	{
		double ans   = 0.0;
		double y0    = f.of(s);
		double x0    = s;

		rand = new Random(123);
		double p_x   = roundDown((s + 1.0));
		double m_x   = roundDown((s - 1.0));

		ans = (d == PLUS_DIRECTION) ? p_x : m_x;
		double y1 = f.of(ans);

		if (nearValue(t, s, y0, ans, y1) == 0) {
			x0 = s;
		} else {
			x0 = ans;
			y0 = y1;
		}

		int i;
		for (i=0; i<loop ;i++) {
			double r = getRatio();
			p_x   = roundDown((ans + r));
			m_x   = roundDown((ans - r));

			double p_y = f.of(p_x);
			double m_y = f.of(m_x);

//logger.debug("AB", ("正:(" + p_x + "," + p_y  + "):負:(" + m_x + "," +  m_y));

			if (nearValue(t, p_x, p_y, m_x, m_y)==0) {
				ans = p_x;
				if (nearValue(t, x0, y0, p_x, p_y) == 1) {
					x0 = p_x;
					y0 = p_y;
				}
			} else {
				ans = m_x;
				if (nearValue(t, x0, y0, m_x, m_y) == 1) {
					x0 = m_x;
					y0 = m_y;
				}
			}
			if (fin(t, y0))	break;
		}

		if (!fin(t, y0)) {
			logger.debug(GoalSeeking.class, (x0 + ":" + y0));
			if (y0 < EPS) {
				throw new UtilMathException(
					UtilMathException.BISECTION_EXCEPTION,
					"値が収束しませんでした。", new Exception());
			} else {
				if (y0 > 5.0e-2) {
					throw new UtilMathException(
						UtilMathException.BISECTION_EXCEPTION,
						"値が収束しませんでした。", new Exception());
				}
			}
		}
		logger.debug(GoalSeeking.class, "#" + (i + 1) + "回で収束しました。"
						+ ":" + x0 + ":" + y0);
		return Math.floor(x0 * 100000.0 + 1.0e-6) / 100000.0;
	}

	private static double roundDown(double x) {
		return Math.floor(x * 100.0 + 1.0e-5) / 100.0;
	}

	private static boolean fin(double t, double y) {
		double a1 = y + (-1 * t);
		return (Math.abs(a1) <= EPS) ? true :  false;
	}

	private static int nearValue(double t, double x1, double y1, double x2, double y2) {
		double a1 = y1 + (-1 * t);
		double a2 = y2 + (-1 * t);
		if (Math.abs(a1) < Math.abs(a2))	return 0;
		else								return 1;
	}
	private static double getRatio() {
		return rand.nextDouble();
	}

}
