package	jp.co.sds_corp.util.net;

import java.io.IOException;
import java.net.InetAddress;
import java.net.SocketException;

/**
* ICMPパケットを送受信するソケットを表します。
*/
public class ICMPSocket extends RawSocket
{
	/**
	* ICMPソケットを構築します。
	*/
	public ICMPSocket() throws SocketException {
		super( 1, null );
	}

	/**
	* ICMPソケットを構築し指定されたアドレスにバインドします。
	* @param	laddr	バインド先のローカルアドレス
	*/
	public ICMPSocket( InetAddress laddr ) throws SocketException {
		super( 1, laddr );
	}

	/**
	* ICMPパケットを送信します。
	* @param	packet	送信するパケット
	*/
	public void send( ICMPPacket packet ) throws IOException  {
		RawPacket raw;
		byte[] buf, aux;

		buf = new byte[packet.getLength() + 8];
		aux = packet.getData();
		for ( int i = 0; i < packet.getLength(); i++ ) {
			buf[i + 8] = aux[i];
		}
		buf[0] = packet.getType().getCode();
		buf[1] = packet.getCode();
		for ( int i = 0; i < 4; i++ ) {
			buf[i+4] = packet.option[i];
		}
		raw = new RawPacket( buf, buf.length, packet.getAddress() );

		switch ( packet.getType() ) {
		case ECHO_REPLY:
		case ECHO_REQUEST:
		case TIMESTAMP_REQUEST:
		case TIMESTAMP_REPLY:
		case INFO_REQUEST:
		case INFO_REPLY:
			raw.buf[4] = (byte )(packet.getIdentifier() & 0xff);
			raw.buf[5] = (byte )((packet.getIdentifier() & 0xff00) >> 8 );
			raw.buf[6] = (byte )(packet.getSequence() & 0xff);
			raw.buf[7] = (byte )((packet.getSequence() & 0xff00) >> 8 );
			break;
		default:
			break;
		}
		raw.buf[2] = 0;
		raw.buf[3] = 0;
		int cs = raw.calcChecksum();
		raw.buf[2] = (byte )(cs & 0xff);
		raw.buf[3] = (byte )((cs & 0xff00) >> 8 );

		super.send( raw );

		packet.setLength( raw.getLength() - 8 );
	}

	/**
	* RawPacketを送信します。
	* @param	p		送信するパケット
	*/
	public void send( RawPacket p ) throws IOException  {
		throw new IOException( "can't send raw packets" );
	}

	/**
	* ICMPパケットを受信します。
	* @param	packet	受信したデータを保存するICMPPacket
	*/
	public synchronized void receive( ICMPPacket packet ) throws IOException {
		byte[] buf;
		int len;
		byte code;
		ICMPPacket.Type type;

		super.receive( packet );

		buf = packet.getData();
		if (( len = packet.getLength()) < 8 ) {
			throw new SocketException( "Not a valid ICMP packet" );
		}
		type  = ICMPPacket.Type.findType( buf[0] );
		code  = buf[1];
		packet.setType( type, code );
		for ( int i = 0; i < 4; i++ ) {
			packet.option[i] = buf[i+4];
		}
		switch ( packet.getType() ) {
		case ECHO_REPLY:
		case ECHO_REQUEST:
		case TIMESTAMP_REQUEST:
		case TIMESTAMP_REPLY:
		case INFO_REQUEST:
		case INFO_REPLY:
			packet.setChecksum( buf[2], buf[3] );
			packet.setIdentifier( buf[4], buf[5] );
			packet.setSequence( buf[6], buf[7] );
			break;
		default:
			break;
		}
		for ( int i = 0; i < len - 8; i++ ) {
			buf[i] = buf[i+8];
		}

		packet.setLength( len - 8 );
	}

	/**
	* このソケットを閉じます。
	*/
	public void close() {
		super.close();
	}

}
